package explorviz.live_trace_processing.filter.reduction.summarization;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

import explorviz.live_trace_processing.filter.IPipeReceiver;
import explorviz.live_trace_processing.filter.reduction.AbstractReductionFilter;
import explorviz.live_trace_processing.record.event.AbstractAfterEventRecord;
import explorviz.live_trace_processing.record.event.AbstractBeforeEventRecord;
import explorviz.live_trace_processing.record.event.AbstractEventRecord;
import explorviz.live_trace_processing.record.trace.Trace;

public class TreeSummarizationFilter extends AbstractReductionFilter {

	public TreeSummarizationFilter(final IPipeReceiver receiver) {
		super(receiver, "TreeSummarization");
	}

	@Override
	protected Trace reduceTrace(final Trace trace) {
		if (trace.getTraceEvents().size() < 4) {
			return trace;
		}
		final List<Edge> edges = new ArrayList<Edge>();
		buildEdgesfromTrace(edges, trace.getTraceEvents());
		final Trace newTrace = buildTraceFromEdges(edges);
		return newTrace;
	}

	private void buildEdgesfromTrace(final List<Edge> edges,
			final List<AbstractEventRecord> traceEvents) {
		AbstractBeforeEventRecord caller;
		AbstractBeforeEventRecord callee = null;
		final Stack<AbstractBeforeEventRecord> callerStack = new Stack<AbstractBeforeEventRecord>();
		callerStack.push(callee);
		final Stack<Integer> stack = new Stack<Integer>();
		final Stack<Edge> edgeStack = new Stack<Edge>();
		int count = 0;
		for (final AbstractEventRecord event : traceEvents) {
			if (event instanceof AbstractBeforeEventRecord) {
				count++;
				caller = callerStack.peek();
				callee = (AbstractBeforeEventRecord) event;
				callerStack.push(callee);
				if (!callerEqualsCallee(caller, callee)
						&& addToList(caller, callee, edges, edgeStack)) {
					stack.push(new Integer(count));
				}
			} else if (event instanceof AbstractAfterEventRecord) {
				callerStack.pop();
				if (!stack.isEmpty() && (count == stack.peek())) {
					stack.pop();
					final Edge edge = edgeStack.pop();
					final AbstractAfterEventRecord callee_end = (AbstractAfterEventRecord) event;
					edge.setCallee_end(callee_end);
				}
				count--;
			} else {
				// TODO: handle remote calls
			}
		}
	}

	private boolean addToList(final AbstractBeforeEventRecord caller,
			final AbstractBeforeEventRecord callee, final List<Edge> edges,
			final Stack<Edge> edgeStack) {
		for (final Edge edge : edges) {
			if (callerEqualsCallee(caller, edge.getCaller())
					&& callerEqualsCallee(callee, edge.getCallee())) {
				mergeInformation(edge, caller, callee);
				return false;
			}
		}
		final Edge edge = new Edge(caller, callee);
		edges.add(edge);
		edgeStack.add(edge);
		return true;
	}

	private void mergeInformation(final Edge edge, final AbstractBeforeEventRecord caller,
			final AbstractBeforeEventRecord callee) {
		edge.getCallee().getRuntimeStatisticInformation()
				.merge(callee.getRuntimeStatisticInformation());
	}

	private Trace buildTraceFromEdges(final List<Edge> edges) {
		final List<AbstractEventRecord> traceEvents = new ArrayList<AbstractEventRecord>();
		final Edge start = edges.get(0);
		traceEvents.add(start.getCallee());
		start.setTaken(true);
		recursiveCall(traceEvents, edges, start.getCallee());
		traceEvents.add(start.getCallee_end());
		return new Trace(traceEvents, true);
	}

	private void recursiveCall(final List<AbstractEventRecord> traceEvents, final List<Edge> edges,
			final AbstractBeforeEventRecord callee) {
		for (final Edge edge : edges) {
			if (!edge.isTaken() && (callerEqualsCallee(edge.getCaller(), callee))) {
				traceEvents.add(edge.getCallee());
				edge.setTaken(true);
				recursiveCall(traceEvents, edges, edge.getCallee());
				traceEvents.add(edge.getCallee_end());
			}
		}

	}

	private boolean callerEqualsCallee(final AbstractBeforeEventRecord caller,
			final AbstractBeforeEventRecord callee) {
		if (caller == null) {
			if (callee == null) {
				return true;
			} else {
				return false;
			}
		} else if (callee == null) {
			return false;
		}
		final String callerClazz = getClazzFullQName(caller.getOperationSignature());
		final String calleeClazz = getClazzFullQName(callee.getOperationSignature());
		return callerClazz.equals(calleeClazz);
	}

	private String getClazzFullQName(final String operationSignatureStr) {
		final int brace = operationSignatureStr.indexOf('(');
		final String signatureWithoutArguments = operationSignatureStr.substring(0, brace);
		final int lastWhiteSpace = signatureWithoutArguments.lastIndexOf(' ');
		final int lastDot = signatureWithoutArguments.lastIndexOf('.');
		if (lastDot == -1) {
			return "";
		}
		final String fullQClazzName = signatureWithoutArguments.substring(lastWhiteSpace + 1,
				lastDot);

		if (fullQClazzName.indexOf("$") > 0) {
			return fullQClazzName.substring(0, fullQClazzName.indexOf("$"));
		} else {
			return fullQClazzName;
		}
	}

}
